<?php

namespace Drupal\graphql_core_schema;

use Drupal\Core\Config\Entity\ConfigEntityType;
use Drupal\Core\Config\TypedConfigManagerInterface;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\Plugin\Field\FieldType\BooleanItem;
use Drupal\Core\Field\Plugin\Field\FieldType\EmailItem;
use Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItem;
use Drupal\Core\Field\Plugin\Field\FieldType\IntegerItem;
use Drupal\Core\Field\Plugin\Field\FieldType\LanguageItem;
use Drupal\Core\Field\Plugin\Field\FieldType\MapItem;
use Drupal\Core\Field\Plugin\Field\FieldType\NumericItemBase;
use Drupal\Core\Field\Plugin\Field\FieldType\StringItem;
use Drupal\Core\Field\Plugin\Field\FieldType\StringItemBase;
use Drupal\Core\Field\Plugin\Field\FieldType\TimestampItem;
use Drupal\Core\Field\TypedData\FieldItemDataDefinition;
use Drupal\Core\Field\TypedData\FieldItemDataDefinitionInterface;
use Drupal\Core\TypedData\DataDefinition;
use Drupal\Core\TypedData\TypedDataManagerInterface;
use Drupal\options\Plugin\Field\FieldType\ListStringItem;
use Drupal\telephone\Plugin\Field\FieldType\TelephoneItem;
use Drupal\text\Plugin\Field\FieldType\TextItemBase;
use Drupal\text\Plugin\Field\FieldType\TextWithSummaryItem;
use GraphQL\Type\Definition\CustomScalarType;
use GraphQL\Type\Definition\EnumType;
use GraphQL\Type\Definition\InterfaceType;
use GraphQL\Type\Definition\ObjectType;
use GraphQL\Type\Definition\Type;
use GraphQL\Type\Schema;
use GraphQL\Utils\SchemaPrinter;

/**
 * The EntitySchemaBuilder class.
 */
class EntitySchemaBuilder extends SchemaBuilder {

  /**
   * Fields that are not resolved by the default field resolver and exist on all entities.
   *
   * @var string[]
   */
  const EXCLUDED_ENTITY_FIELDS = [
    'id',
    'uuid',
    'label',
    'langcode',
  ];

  /**
   * The enabled entity types.
   *
   * @var string[]
   */
  protected $enabledEntityTypes;

  /**
   * Enabled extensions.
   *
   * @var string[]
   */
  protected $enabledExtensions = [];

  /**
   * Enabled fields.
   *
   * @var string[][]
   */
  protected $enabledFields = [];

  /**
   * Generate value fields.
   *
   * @var bool
   */
  protected $generateValueFields;

  /**
   * The constructor.
   *
   * @param \Drupal\Core\TypedData\TypedDataManagerInterface $typedDataManager
   *   The typed data manager.
   * @param \Drupal\Core\Config\TypedConfigManagerInterface $typedConfigManager
   *   The typed config manager.
   * @param array $enabledEntityTypes
   *   The enabled entity types.
   * @param array $enabledExtensions
   *   The enabled extensions.
   * @param array $enabledFields
   *   List of enabled fields.
   */
  public function __construct(
    TypedDataManagerInterface $typedDataManager,
    TypedConfigManagerInterface $typedConfigManager,
    array $enabledEntityTypes,
    array $enabledExtensions,
    array $enabledFields
  ) {
    parent::__construct($typedDataManager, $typedConfigManager);

    $this->generateValueFields = TRUE;
    $this->enabledExtensions = $enabledExtensions;
    $this->enabledEntityTypes = $enabledEntityTypes;
    $this->enabledFields = $enabledFields;

    // Create scalars.
    $this->types['MapData'] = new CustomScalarType([
      'name' => 'MapData',
    ]);

    // Create entity interfaces.
    $this->createInterface('Entity', 'A common interface for all entity objects.', [
      'id' => [
        'type' => Type::nonNull(Type::string()),
      ],
      'uuid' => [
        'type' => Type::nonNull(Type::string()),
        'description' => 'The unique UUID.',
      ],
      'label' => [
        'type' => Type::string(),
        'description' => 'The label of this entity.',
      ],
      'langcode' => [
        'type' => Type::string(),
        'description' => 'The langcode of this entity.',
      ],
      'toArray' => [
        'type' => fn() => $this->types['MapData'],
        'description' => 'Gets an array of all property values.',
      ],
      'getConfigTarget' => [
        'type' => Type::string(),
        'description' => 'Gets the configuration target identifier for the entity.',
      ],
      'uriRelationships' => [
        'type' => Type::listOf(Type::string()),
        'description' => 'Gets a list of URI relationships supported by this entity.',
      ],
      'referencedEntities' => [
        'type' => fn () => Type::listOf($this->types['Entity']),
        'description' => 'Gets a list of entities referenced by this entity.',
      ],
      'entityTypeId' => [
        'type' => Type::nonNull(Type::string()),
        'description' => 'The entity type ID.',
      ],
      'isNew' => [
        'type' => Type::nonNull(Type::boolean()),
        'description' => 'Determines whether the entity is new.',
      ],
      'accessCheck' => [
        'type' => Type::nonNull(Type::boolean()),
        'args' => [
          'operation' => Type::string(),
        ],
        'description' => 'Check entity access for the given operation, defaults to view.',
      ],
    ]);
    $this->createInterface('EntityTranslatable', 'An entity that is translatable.', [
      'translations' => [
        'type' => fn() => Type::listOf($this->types['EntityTranslatable']),
        'description' => 'Get all translations.',
      ],
      'translation' => [
        'type' => fn() => $this->types['EntityTranslatable'],
        'args' => [
          'langcode' => fn() => Type::nonNull($this->types['Langcode']),
        ],
        'description' => 'Get a specific translation.',
      ],
    ]);
    $this->createInterface('EntityLinkable', 'An entity that is linkable.', [
      'url' => [
        'type' => fn() => $this->types['Url'],
        'args' => [
          'rel' => Type::string(),
        ],
        'description' => 'Get the URL, defaults to canonical.',
      ],
    ]);

    // Create field item and field item list interfaces.
    $baseFieldItemListFields = [
      'fieldConfig' => function () {
        return fn() => Type::nonNull($this->types['FieldDefinition']);
      },
      'isEmpty' => [
        'type' => fn() => Type::nonNull(Type::boolean()),
        'description' => 'True if the field list has no items.',
      ],
      'count' => [
        'type' => fn() => Type::nonNull(Type::int()),
        'description' => 'The number of field items.',
      ],
      'getString' => [
        'type' => fn() => Type::nonNull(Type::string()),
        'description' => 'Get a string representation of all field items.',
      ],
    ];

    $baseFieldItemFields = [
      'isEmpty' => [
        'type' => Type::nonNull(Type::boolean()),
        'description' => 'True if this item is considered empty.',
      ],
    ];

    if ($this->extensionEnabled('render_field_item')) {
      $baseFieldItemFields['viewFieldItem'] = [
        'type' => Type::nonNull(Type::string()),
        'args' => [
          'viewMode' => Type::string(),
        ],
        'description' => 'Render the field item in the given viewmode.',
      ];
      $baseFieldItemListFields['viewField'] = [
        'type' => Type::nonNull(Type::string()),
        'args' => [
          'viewMode' => Type::string(),
        ],
        'description' => 'Render the field in the given viewmode.',
      ];
    }

    $this->createInterface('FieldItemList', 'A field item list containing items.', $baseFieldItemListFields);
    $this->createInterface('FieldItemType', 'An item in a field list.', $baseFieldItemFields);

    // Create interface for the field definition.
    $this->createInterface('FieldDefinition', 'Interface for field definitions.', [
      'name' => Type::string(),
    ]);

    // Interface for an URL.
    $this->createInterface('Url', 'Interface for an URL.', [
      'path' => Type::string(),
    ]);

    $this->createInterface('LanguageInterface', 'Interface for a language.', [
      'name' => Type::nonNull(Type::string()),
      'id' => Type::nonNull(Type::string()),
      'direction' => Type::nonNull(Type::int()),
      'weight' => Type::nonNull(Type::int()),
      'isLocked' => Type::nonNull(Type::boolean()),
    ]);

    $this->types['Language'] = new ObjectType([
      'name' => 'Language',
      'interfaces' => [$this->types['LanguageInterface']],
      'fields' => $this->types['LanguageInterface']->getFields(),
    ]);
  }

  /**
   * Check if the given field is enabled.
   *
   * @param string $entityTypeId
   *   The entity type ID.
   * @param string $fieldName
   *   The field name.
   *
   * @return bool
   *   TRUE if the field is enabled for this entity type.
   */
  private function fieldIsEnabled(string $entityTypeId, string $fieldName) {
    if (empty($this->enabledFields[$entityTypeId])) {
      return FALSE;
    }
    return in_array($fieldName, $this->enabledFields[$entityTypeId]);
  }

  /**
   * Check if the given extension is enabled.
   *
   * @param string $id
   *   The ID of the extension.
   *
   * @return bool
   *   The extension is enabled.
   */
  private function extensionEnabled(string $id): bool {
    return in_array($id, $this->enabledExtensions);
  }

  /**
   * Check if the given entity type is enabled.
   *
   * @param string $id
   *   The ID of the entity type.
   *
   * @return bool
   *   The entity type is enabled.
   */
  private function entityTypeIsEnabled(string $id): bool {
    return in_array($id, $this->enabledEntityTypes);
  }

  /**
   * Get the GraphQL field definition for an entity value field.
   *
   * This will return a type for value fields, where instead of the entire
   * field the direct field value is resolved. For example a simple text
   * field will directly resolve to a string scalar.
   *
   * If no appropriate scalar is found the type for the field item is returned.
   *
   * @param \Drupal\Core\Field\FieldDefinitionInterface $fieldDefinition
   *   The field definition.
   *
   * @return \GraphQL\Type\Definition\Type|array|null
   *   The GraphQL type if found.
   */
  private function buildGraphqlValueField(FieldDefinitionInterface $fieldDefinition) {
    $description = (string) $fieldDefinition->getDescription();
    $fieldName = $fieldDefinition->getName();
    if (!$description) {
      $description = (string) $fieldDefinition->getFieldStorageDefinition()->getDescription();
    }
    $description .= ' (field value)';
    $storageDefinition = $fieldDefinition->getFieldStorageDefinition();
    $isMultiple = $storageDefinition->isMultiple();

    // Create a field item we can use to determine what scalar this should
    // resolve to.
    $itemDefinition = $fieldDefinition->getItemDefinition();
    $typedData = $this->typedDataManager->create($itemDefinition);

    if ($fieldName === 'metatag' && $typedData instanceof MapItem) {
      return NULL;
    }

    if (
      $typedData instanceof StringItem ||
      $typedData instanceof StringItemBase ||
      $typedData instanceof EmailItem ||
      $typedData instanceof TelephoneItem
    ) {
      return [
        'type' => $this->wrapTypeList(Type::string(), $isMultiple),
        'description' => $description,
      ];
    }
    elseif ($typedData instanceof LanguageItem) {
      return [
        'type' => $this->wrapTypeList($this->types['LanguageInterface'], $isMultiple),
        'description' => $description,
      ];
    }
    elseif ($typedData instanceof ListStringItem) {
      return [
        'type' => Type::listOf(Type::string()),
        'description' => $description,
      ];
    }
    elseif ($typedData instanceof IntegerItem) {
      return [
        'type' => $this->wrapTypeList(Type::int(), $isMultiple),
        'description' => $description,
      ];
    }
    elseif ($typedData instanceof NumericItemBase) {
      return [
        'type' => $this->wrapTypeList(Type::float(), $isMultiple),
        'description' => $description,
      ];
    }
    elseif ($typedData instanceof BooleanItem) {
      return [
        'type' => $this->wrapTypeList(Type::boolean(), $isMultiple),
        'description' => $description,
      ];
    }
    elseif ($typedData instanceof EntityReferenceItem) {
      $type = $this->getTypeForEntityReferenceFieldItem($itemDefinition, $isMultiple);
      if ($type) {
        return [
          'type' => fn() => $this->wrapTypeList($this->types[$type], $isMultiple),
          'description' => $description,
        ];
      }
      // The entity type that is referenced is not enabled, so we don't output
      // this field at all.
      return NULL;
    }
    elseif ($typedData instanceof TextWithSummaryItem) {
      return [
        'type' => $this->wrapTypeList(Type::string(), $isMultiple),
        'description' => $description,
        'args' => [
          'summary' => Type::boolean(),
        ],
      ];
    }
    elseif ($typedData instanceof TextItemBase) {
      return [
        'type' => $this->wrapTypeList(Type::string(), $isMultiple),
        'description' => $description,
      ];
    }
    elseif ($typedData instanceof TimestampItem) {
      return [
        'type' => $this->wrapTypeList(Type::string(), $isMultiple),
        'description' => $description,
      ];
    }

    // The field type is not scalar, try to get the GraphQL type for this item
    // definition.
    $itemType = $this->getFieldItemType($itemDefinition);
    if ($itemType) {
      return [
        'type' => $this->wrapTypeList($itemType, $isMultiple),
        'description' => $description,
      ];
    }

    return NULL;
  }

  /**
   * Generate the GraphQL type for an entity reference field item.
   *
   * @param \Drupal\Core\Field\TypedData\FieldItemDataDefinitionInterface $itemDefinition
   *   The field definition.
   *
   * @return string|null
   *   The name of the referenced type.
   */
  private function getTypeForEntityReferenceFieldItem(FieldItemDataDefinitionInterface $itemDefinition): ?string {
    $targetType = $itemDefinition->getSetting('target_type');
    $handlerSettings = $itemDefinition->getSetting('handler_settings') ?? [];
    $targetBundles = $handlerSettings['target_bundles'] ?? [];
    if (is_string($targetBundles)) {
      $targetBundles = [$targetBundles];
    }
    $targetBundles = array_values($targetBundles);

    // Check if the target entity type is enabled.
    if (!in_array($targetType, $this->enabledEntityTypes)) {
      return NULL;
    }

    if (!empty($targetBundles)) {
      // Any or more than 1 target bundles allowed. The field type will be the
      // entity type.
      if (count($targetBundles) > 1) {
        return EntitySchemaHelper::toPascalCase([$targetType]);
      }

      // Only a single bundle is allowed. The field type will be this specific
      // bundle.
      return EntitySchemaHelper::toPascalCase([
        $targetType,
        $targetBundles[0],
      ]);
    }

    if ($targetType) {
      return EntitySchemaHelper::toPascalCase([$targetType]);
    }

    return NULL;
  }

  /**
   * Build the FieldItemList type for a field type.
   *
   * @param \Drupal\Core\Field\FieldDefinitionInterface $fieldDefinition
   *   The field definition.
   *
   * @return \GraphQL\Type\Definition\Type|null
   *   The GraphQL type.
   */
  private function getFieldItemListType(FieldDefinitionInterface $fieldDefinition): ?Type {
    $fieldTypeName = $fieldDefinition->getType();
    $graphqlTypeName = EntitySchemaHelper::toPascalCase([
      'field_item_list_',
      $fieldTypeName,
    ]);

    // Type was already generated.
    if (!empty($this->types[$graphqlTypeName])) {
      return $this->types[$graphqlTypeName];
    }

    $fields = [
      ...$this->types['FieldItemList']->getFields(),
    ];

    $itemDefinition = $fieldDefinition->getItemDefinition();
    if ($itemDefinition instanceof FieldItemDataDefinition) {
      $fieldItemType = $this->getFieldItemType($itemDefinition);
      if ($fieldItemType) {
        $fields['list'] = [
          'type' => fn() => Type::listOf($fieldItemType),
          'description' => 'Array of field items.',
        ];
        $fields['first'] = [
          'type' => fn() => $fieldItemType,
          'description' => 'The first field item.',
        ];
      }
    }
    if (empty($fields)) {
      return NULL;
    }

    $type = [
      'name' => $graphqlTypeName,
      'interfaces' => [
        fn() => $this->types['FieldItemList'],
      ],
      'fields' => $fields,
      'description' => (string) $fieldDefinition->getLabel(),
    ];

    $objectType = new ObjectType($type);
    $this->types[$graphqlTypeName] = $objectType;
    return $this->types[$graphqlTypeName];
  }

  /**
   * Check if the given entity reference field should be added.
   */
  private function shouldAddField(string $fieldName, FieldDefinitionInterface $definition) {
    if (in_array($fieldName, self::EXCLUDED_ENTITY_FIELDS)) {
      return FALSE;
    }
    $fieldType = $definition->getType();
    if (in_array($fieldType, self::EXCLUDED_TYPES)) {
      return FALSE;
    }
    if ($fieldType === 'entity_reference' || $fieldType === 'entity_reference_revisions') {
      $targetType = $definition->getSetting('target_type');
      return $this->entityTypeIsEnabled($targetType);
    }

    return TRUE;
  }

  /**
   * Build the type for a field item definition.
   *
   * @param \Drupal\Core\Field\TypedData\FieldItemDataDefinitionInterface $itemDefinition
   *   The field item data definition.
   *
   * @return \GraphQL\Type\Definition\Type|null
   *   The GraphQL type if available.
   */
  private function getFieldItemType(FieldItemDataDefinitionInterface $itemDefinition): ?Type {
    $type = $itemDefinition->getFieldDefinition()->getType();
    $graphqlDataTypeName = EntitySchemaHelper::toPascalCase(
      ['field_item_type_', $type]
    );
    if (!empty($this->types[$graphqlDataTypeName])) {
      return $this->types[$graphqlDataTypeName];
    }

    $fields = [];
    $propertyDefinitions = $itemDefinition->getPropertyDefinitions();

    foreach ($propertyDefinitions as $name => $propertyDefinition) {
      if ($propertyDefinition instanceof DataDefinition) {
        $propertyFieldType = $this->getDataPropertyType($propertyDefinition->toArray());
        if ($propertyFieldType) {
          $propertyFieldName = EntitySchemaHelper::toCamelCase($name);
          $fields[$propertyFieldName] = [
            'type' => fn() => $propertyFieldType,
            'description' => (string) $propertyDefinition->getLabel(),
          ];
        }
      }
    }

    if (empty($fields)) {
      return NULL;
    }

    $type = new ObjectType([
      'name' => $graphqlDataTypeName,
      'interfaces' => [
        fn() => $this->types['FieldItemType'],
      ],
      'fields' => [...$fields, ...$this->types['FieldItemType']->getFields()],
      'description' => (string) $itemDefinition->getLabel(),
    ]);

    $this->types[$graphqlDataTypeName] = $type;
    return $this->types[$graphqlDataTypeName];
  }

  /**
   * Add types for the entity type.
   *
   * @param \Drupal\Core\Entity\EntityTypeInterface $entityType
   *   The entity type.
   * @param \Drupal\Core\Field\FieldDefinitionInterface[] $fieldDefinitions
   *   The base field definitions of the entity type.
   */
  public function addContentEntityType(EntityTypeInterface $entityType, array $fieldDefinitions) {
    $typeName = EntitySchemaHelper::toPascalCase([$entityType->id()]);
    $hasBundles = $entityType->hasKey('bundle');
    $isLinkable = !empty($entityType->getLinkTemplates());

    $fields = [
      ...$this->types['Entity']->getFields(),
      ...$this->createEntityFields($entityType->id(), $fieldDefinitions),
    ];
    $interfaces = [
      $this->types['Entity'],
    ];
    if ($isLinkable) {
      $interfaces[] = $this->types['EntityLinkable'];
      $fields = array_merge($fields, $this->types['EntityLinkable']->getFields());
    }

    $config = [
      'name' => $typeName,
      'interfaces' => $interfaces,
      'fields' => $fields,
      'description' => (string) $entityType->getLabel(),
    ];

    $type = $hasBundles ? new InterfaceType($config) : new ObjectType($config);
    $this->types[$typeName] = $type;
  }

  /**
   * Add types for the configuration entity type.
   *
   * @param \Drupal\Core\Config\Entity\ConfigEntityType $entityType
   *   The config entity type ID.
   * @param array $mapping
   *   The schema mapping for the config entity.
   */
  public function addConfigEntityType(ConfigEntityType $entityType, array $mapping) {
    $entityTypeId = $entityType->id();
    $typeName = EntitySchemaHelper::toPascalCase([$entityTypeId]);

    $fields = [
      ...$this->types['Entity']->getFields(),
    ];

    foreach ($mapping as $propertyName => $definition) {
      $graphqlFieldName = EntitySchemaHelper::toCamelCase($propertyName);
      if (in_array($propertyName, self::EXCLUDED_ENTITY_FIELDS)) {
        continue;
      }
      if (!$this->fieldIsEnabled($entityTypeId, $propertyName)) {
        continue;
      }

      $type = $this->getDataPropertyType($definition);
      if ($type) {
        $fields[$graphqlFieldName] = [
          'type' => fn() => $type,
        ];
      }
    }
    $interfaces = [
      $this->types['Entity'],
    ];

    if ($entityTypeId === 'field_config') {
      $interfaces[] = $this->types['FieldDefinition'];
      $fields = array_merge($fields, $this->types['FieldDefinition']->getFields());
    }
    elseif ($entityTypeId === 'configurable_language') {
      $interfaces[] = $this->types['LanguageInterface'];
      $fields = array_merge($fields, $this->types['LanguageInterface']->getFields());
    }

    $type = new ObjectType([
      'name' => $typeName,
      'interfaces' => $interfaces,
      'fields' => $fields,
      'description' => (string) $entityType->getLabel(),
    ]);

    $this->types[$typeName] = $type;
  }

  /**
   * Add types for the entity bundle type.
   *
   * @param \Drupal\Core\Entity\ContentEntityTypeInterface $entityType
   *   The entity type.
   * @param string $bundleId
   *   The bundle ID.
   * @param array $bundleInfo
   *   The bundle info.
   * @param \Drupal\Core\Field\FieldDefinitionInterface[] $fieldDefinitions
   *   The field definitions of the bundle.
   */
  public function addContentEntityBundleType(ContentEntityTypeInterface $entityType, string $bundleId, array $bundleInfo, array $fieldDefinitions) {
    $entityTypeId = $entityType->id();
    $entityTypeName = EntitySchemaHelper::toPascalCase([$entityTypeId]);
    $bundleTypeName = EntitySchemaHelper::toPascalCase(
      [$entityTypeId, $bundleId]
    );
    $fields = [
      ...$this->types['Entity']->getFields(),
      ...$this->types[$entityTypeName]->getFields(),
      ...$this->createEntityFields($entityType->id(), $fieldDefinitions),
    ];
    $interfaces = [
      $this->types['Entity'],
      $this->types[$entityTypeName],
    ];

    // An entity is linkable when it defines at least one link template (e.g.
    // canonical).
    if (!empty($entityType->getLinkTemplates())) {
      $interfaces[] = $this->types['EntityLinkable'];
      $fields = array_merge($fields, $this->types['EntityLinkable']->getFields());
    }

    // Add the interface and fields for a translatable entity.
    if (!empty($bundleInfo['translatable'])) {
      $interfaces[] = $this->types['EntityTranslatable'];
      $fields['translations'] = [
        'type' => fn() => Type::listOf($this->types[$bundleTypeName]),
      ];
      $fields['translation'] = [
        'type' => fn() => $this->types[$bundleTypeName],
        'args' => [
          'langcode' => fn() => Type::nonNull($this->types['Langcode']),
        ],
      ];
    }

    $type = new ObjectType([
      'name' => $bundleTypeName,
      'interfaces' => $interfaces,
      'fields' => $fields,
      'description' => (string) $bundleInfo['label'] ?? NULL,
    ]);

    $this->types[$bundleTypeName] = $type;
  }

  /**
   * Create GraphQL fields given the entity field definitions.
   *
   * @param string $entityTypeId
   *   The entity type ID the fields belong to.
   * @param \Drupal\Core\Field\FieldDefinitionInterface[] $fieldDefinitions
   *   The field definitions.
   *
   * @return array
   *   The array of GraphQL fields.
   */
  private function createEntityFields(string $entityTypeId, array $fieldDefinitions): array {
    $fields = [];
    foreach ($fieldDefinitions as $fieldName => $definition) {
      // Try to get the type first. This way we can make sure that we generate
      // a type for every field type, even if no entity has a field with that
      // type.
      $type = $this->getFieldItemListType($definition);

      if (!$this->fieldIsEnabled($entityTypeId, $fieldName)) {
        continue;
      }
      if (!$this->shouldAddField($fieldName, $definition)) {
        continue;
      }
      if ($type) {
        $graphqlFieldName = EntitySchemaHelper::toCamelCase(
          [$fieldName, '_raw_field']
        );
        $fields[$graphqlFieldName] = [
          'type' => $type,
          'description' => (string) $definition->getDescription(),
        ];
      }
      if ($this->generateValueFields) {
        $valueField = $this->buildGraphqlValueField($definition);
        if ($valueField) {
          $valueFieldName = EntitySchemaHelper::toCamelCase($fieldName);
          $fields[$valueFieldName] = $valueField;
        }
      }
    }

    return $fields;
  }

  /**
   * Create the enum for the date formats.
   *
   * @param \Drupal\Core\Datetime\Entity\DateFormat[] $formats
   *   The date format entities.
   */
  public function addDateFormatEnum(array $formats) {
    /** @var \Drupal\Core\Datetime\DateFormatterInterface $dateFormatter */
    $dateFormatter = \Drupal::service('date.formatter');
    $timestamp = 1668433941;

    $values = [];
    foreach ($formats as $format) {
      $id = (string) $format->id();
      $key = strtoupper(EntitySchemaHelper::toSnakeCase($id));
      $description = $dateFormatter->format($timestamp, $id);
      $values[$key] = [
        'value' => $id,
        'description' => $description,
      ];
    }
    $enum = new EnumType([
      'name' => 'DrupalDateFormat',
      'values' => $values,
    ]);
    $this->types['DrupalDateFormat'] = $enum;
  }

  /**
   * Create the enum for the entity types.
   *
   * @param \Drupal\Core\Entity\EntityType[] $types
   *   The entity type definitions.
   */
  public function addEntityTypeEnum(array $types) {
    $values = [];
    foreach ($types as $type) {
      $key = (string) $type->id();
      if (in_array($key, $this->enabledEntityTypes)) {
        $key = strtoupper(EntitySchemaHelper::toSnakeCase($key));
        $values[$key] = [
          'value' => $type->id(),
          'description' => (string) $type->getLabel(),
        ];
      }
    }
    $enum = new EnumType([
      'name' => 'EntityType',
      'values' => $values,
    ]);
    $this->types['EntityTypeEnum'] = $enum;
  }

  /**
   * Create the enum for the langcodes.
   *
   * @param \Drupal\Core\Language\Language[] $languages
   *   The entity type definitions.
   */
  public function addLangcodeEnum(array $languages) {
    $values = [];
    foreach ($languages as $language) {
      $key = (string) $language->getId();
      $key = strtoupper(EntitySchemaHelper::toSnakeCase($key));
      $values[$key] = [
        'value' => $language->getId(),
        'description' => $language->getName(),
      ];
    }
    $enum = new EnumType([
      'name' => 'Langcode',
      'values' => $values,
    ]);
    $this->types['Langcode'] = $enum;
  }

  /**
   * Get the generated schema.
   *
   * @return string
   *   The generated schema.
   */
  public function getGeneratedSchema() {
    $schema = new Schema([
      'types' => $this->types,
    ]);

    $result = SchemaPrinter::doPrint($schema);
    return $result;
  }

}
