<?php

namespace Drupal\graphql_core_schema;

use Drupal\Core\Config\TypedConfigManagerInterface;
use Drupal\Core\TypedData\TypedDataManagerInterface;
use GraphQL\Type\Definition\InterfaceType;
use GraphQL\Type\Definition\ObjectType;
use GraphQL\Type\Definition\Type;
use GraphQL\Type\Schema;
use GraphQL\Utils\SchemaPrinter;

/**
 * The SchemaBuilder base class.
 */
abstract class SchemaBuilder {

  /**
   * Types that should never be generated.
   *
   * @var string[]
   */
  const EXCLUDED_TYPES = [
    'password',
    '_core_config_info',
  ];

  /**
   * The type data manager.
   *
   * @var \Drupal\Core\TypedData\TypedDataManager
   */
  protected $typedDataManager;

  /**
   * The typed config manager.
   *
   * @var \Drupal\Core\Config\TypedConfigManagerInterface
   */
  protected $typedConfigManager;

  /**
   * The generated schema.
   *
   * @var \GraphQL\Type\Definition\TypeWithFields[]
   */
  protected $types;

  /**
   * Array of types that have been or will be generated.
   *
   * Unlike the $types property, this array may contain a type that is in the
   * process of being generated, e.g. that the type can already be referenced
   * using a callable, since it will be available once the callables are
   * called.
   *
   * @var string[]
   */
  protected $generatedTypeNames;

  /**
   * The constructor.
   *
   * @param \Drupal\Core\TypedData\TypedDataManagerInterface $typedDataManager
   *   The typed data manager.
   * @param \Drupal\Core\Config\TypedConfigManagerInterface $typedConfigManager
   *   The typed config manager.
   */
  public function __construct(
    TypedDataManagerInterface $typedDataManager,
    TypedConfigManagerInterface $typedConfigManager
  ) {
    $this->typedDataManager = $typedDataManager;
    $this->typedConfigManager = $typedConfigManager;
    $this->generatedTypeNames = [];
    $this->types = [];
  }

  /**
   * Create the needed interfaced.
   *
   * @param string $name
   *   The name of the interface.
   * @param string $description
   *   The description.
   * @param array $fields
   *   The fields.
   */
  protected function createInterface(string $name, string $description, array $fields) {
    $this->types[$name] = new InterfaceType([
      'name' => $name,
      'fields' => $fields,
      'description' => $description,
    ]);
  }

  /**
   * Helper to wrap a type in a list type (when it should resolve to an array).
   *
   * @param \GraphQL\Type\Definition\Type $type
   *   The GraphQL type.
   * @param bool $isList
   *   TRUE if the type should be wrapped as a list.
   *
   * @return \GraphQL\Type\Definition\Type
   *   The type, wrapped as a list type if desired.
   */
  protected function wrapTypeList(Type $type, bool $isList): Type {
    if ($isList && $type) {
      return Type::listOf($type);
    }
    return $type;
  }

  /**
   * Get the GraphQL type for a data property definition.
   *
   * This is the lowest possible leaf in the entity schema. It usually resolves
   * to a scalar, but special handling is implemented for sequence types and
   * entity reference types.
   *
   * @param array $definition
   *   The property definition.
   *
   * @return \GraphQL\Type\Definition\Type|null
   *   The GraphQL type.
   */
  protected function getDataPropertyType(array $definition) {
    $type = $definition['type'];

    if (in_array($type, self::EXCLUDED_TYPES)) {
      return NULL;
    }

    // Basic types.
    switch ($type) {
      case 'string':
      case 'email':
      case 'text':
      case 'label':
      case 'path':
      case 'color_hex':
      case 'date_form':
      case 'filter_format':
      case 'datetime_iso8601':
      case 'timestamp':
        return Type::string();

      case 'boolean':
        return Type::boolean();

      case 'integer':
        return Type::int();

      case 'float':
        return Type::float();

      case 'uri':
        return $this->types['Url'];

      case 'config_dependencies':
        return $this->types['MapData'];

      case '_core_config_info':
        return NULL;

      case 'entity_reference':
      case 'entity_revision_reference':
        return $this->types['Entity'];

      case 'language_reference':
        return $this->types['LanguageInterface'];
    }

    // A sequence is an array of items.
    if ($type === 'sequence' && !empty($definition['sequence']['type'])) {
      // Try to find a possible type for the sequence.
      $sequenceType = $this->getDataPropertyType($definition['sequence']);
      if ($sequenceType) {
        return Type::listOf($sequenceType);
      }
      return NULL;
    }

    // Try to find a matching data definition for this type.
    if ($this->typedDataManager->hasDefinition($type)) {
      $dataDefinition = $this->typedDataManager->getDefinition($type);
    }
    elseif ($this->typedConfigManager->hasDefinition($type)) {
      $dataDefinition = $this->typedConfigManager->getDefinition($type);
    }

    // A mapping is basically a type that references another type.
    // The method being called here will eventually call this method again. If
    // the referenced map type again references a map type, it might end up
    // here a third time and so on. In the end we have eventually resolved to a
    // scalar type being returned above.
    // This allows us to fully resolve config schema types down to the last
    // property, if supported.
    if ($dataDefinition && !empty($dataDefinition['mapping'])) {
      return $this->getTypeForMapping($type, $dataDefinition['mapping']);
    }

    return NULL;
  }

  /**
   * Get the GraphQL type name for a schema type mapping.
   *
   * @param string $mappingName
   *   The name of the mapping.
   *
   * @return string
   *   The GraphQL type name.
   */
  protected function getGraphqlTypeNameForMapping(string $mappingName) {
    return EntitySchemaHelper::toPascalCase([
      'data_type_',
      $mappingName,
    ]);
  }

  /**
   * Try to infer the type for a mapping property.
   *
   * @param string $mappingName
   *   The name of the mapping.
   * @param array $mapping
   *   The mapping configuration.
   *
   * @return \GraphQL\Type\Definition\Type|null
   *   The GraphQL type if found.
   */
  private function getTypeForMapping(string $mappingName, array $mapping) {
    // E.g. DataTypeLinkitMatcher.
    $graphqlTypeName = $this->getGraphqlTypeNameForMapping($mappingName);

    // Type already generated.
    if (!empty($this->types[$graphqlTypeName])) {
      return $this->types[$graphqlTypeName];
    }
    // Type is being generated in this recursive callstack. Assume it will be
    // there when the schema is being resolved.
    if (in_array($graphqlTypeName, $this->generatedTypeNames)) {
      return fn() => $this->types[$graphqlTypeName];
    }

    // Store this type name in a special variable to make sure we don't end up
    // in an infinite loop.
    $this->generatedTypeNames[] = $graphqlTypeName;

    $mappingFields = [];
    foreach ($mapping as $mappingProperty => $mappingDefinition) {
      $mappingType = $this->getDataPropertyType($mappingDefinition);
      if ($mappingType) {
        $mappingFields[$mappingProperty] = $mappingType;
      }
    }
    if (!empty($mappingFields)) {
      $this->types[$graphqlTypeName] = new ObjectType([
        'name' => $graphqlTypeName,
        'fields' => $mappingFields,
        'description' => "The $mappingName schema mapping.",
      ]);
      return $this->types[$graphqlTypeName];
    }

    // Fallback to the MapData scalar so that the data is at least somehow
    // available.
    return $this->types['MapData'];
  }

  /**
   * Get the generated schema.
   *
   * @return string
   *   The generated schema.
   */
  public function getGeneratedSchema() {
    $schema = new Schema([
      'types' => $this->types,
    ]);

    $result = SchemaPrinter::doPrint($schema);
    return $result;
  }

}
